library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;

entity sdram_ctrl is
generic(
	COL_WIDTH  : integer                       := 8        ;--2^8 = 256 addresses in each colum
	ROW_WIDTH  : integer                       := 12       ;--2^12 = 4096 addresses in each bank
	BANK_WIDTH : integer                       := 2        ;--2^2 = 4 banks in a single chip
	DQ_WIDTH   : integer                       := 8        ;--8 bit Data Bus for one chip
	SDRAM_MR   : std_logic_vector(15 downto 0) := X"0037"   --Full Page Burst,Latency=3,Burst Read and Burst Write,Standard mode
);
port(
	clk   : in std_logic;--The main clock of the control system
	rst_n : in std_logic;--A low voltage may bring the control system to an oringinal state,but the data in SDRAM may be damaged
	
	user_command_req  : in     std_logic                                                    ;--The user want to send a read/write command,active HIGH
	user_command_gnt  : buffer std_logic                                                    ;--The state machine have accept the user's read/write command
	user_rd0_wr1      : in     std_logic                                                    ;--HIGH voltage means read and a low means write,synchronous to "user_command_req"
	user_burst_length : in     std_logic_vector((COL_WIDTH-1) downto 0)                     ;--Express the read/write lenghth of each command,synchronous to "user_command_req"
	user_start_addr   : in     std_logic_vector((BANK_WIDTH+ROW_WIDTH+COL_WIDTH-1) downto 0);--the user's read/write address,synchronous to "user_command_req"
	Tx_fifo_rd_en     : buffer std_logic                                                    ;--used in the user's write command,this port is synchronous to "clk0"
	Tx_fifo_rd_data   : in     std_logic_vector((DQ_WIDTH/8+DQ_WIDTH-1) downto 0)           ;--the user write data to the fifo,synchronous to "clk0"
	
	sdram_cke       : out std_logic                                ;
	sdram_cs_n      : out std_logic                                ;
	sdram_ras_n     : out std_logic                                ;
	sdram_cas_n     : out std_logic                                ;
	sdram_we_n      : out std_logic                                ;
	sdram_ba        : out std_logic_vector((BANK_WIDTH-1) downto 0);
	sdram_addr      : out std_logic_vector((ROW_WIDTH-1) downto 0) ;
	sdram_dq_out_en : out std_logic                                ;
	sdram_dqm       : out std_logic_vector((DQ_WIDTH/8-1) downto 0);
	sdram_dq_out    : out std_logic_vector((DQ_WIDTH-1)   downto 0);
	sdram_dq_in     : in  std_logic_vector((DQ_WIDTH-1)   downto 0);
	
	sdram_data_valid : out std_logic                              ;--tell the user that valid read data is coming,active HIGH
	sdram_rd_data    : out std_logic_vector((DQ_WIDTH-1) downto 0) --synchronous to sdram_data_valid
);
end sdram_ctrl;

	--1.ģ(sdram_ctrl.v)SDRAMĺĿ߼clkΪοʱӡһģ(sdram_io.v)ҪǶSDRAMIOƽӿڵʱ
	--λλľƫΧFPGAڲߺźŵӳ١PCBԼSDRAMĲйأûԸʵ
	
	--2.user_command_reqʱʾûҪִһдuser_command_reqźӦһֱЧ
	--user_command_gntЧģuser_command_gntЧʱuser_rd0_wr1user_DQMuser_burst_length
	--user_start_addrԾʲôSDRAM
	
	--3.SDRAMæڳʼʱִûˢʱûκֻеuser_command_gntЧ
	--ЧʱűʾģѾûĶ/д
	
	--4.ûǶuser_rd0_wr1user_command_reqЧʱǵ͵ƽд֮
	
	--5.user_burst_lengthʾǶSDRAMͻдĳȡ0ʾͻΪ1511ʾͻ
	--512Դơ
	
	--6.SDRAMͻд󳤶2^COL_WIDTH(2COL_WIDTHη)ͻдʱһҪرͻĵַΧҪԽ硣
	--ԽָһͻУͻʼַͽַڲͬROWСԽ磬ôʵͻַعǰҳʼַ
	--һҳʼַ
	
	--7.user_start_addrָǱβʼַ
	
	--8.дûҪȽдSDRAMݰ˳дTx_fifoСTx_fifoⲿģṩģTx_fifoж
	--ûǱߵٶ죬ҲȷдٽдݴfifoΪģڽյûҪȴ6
	--ʱڲſܷfifoеݡfifoĬΪ1Pageݡ
	
	--9.ÿһͻд֮ǰȽfifoգȻڽдfifo
	
	--10.Ƕûڷ󣬻sdram_data_validźšsdram_data_validЧʱʾ
	--SDRAMжѾءͻΪnsdram_data_validЧnʱڡsdram_rd_dataҲ
	--ʱڵƶ仯ʾǴSDRAMжݡ1ݶӦͻʼַݣn
	--ݶӦͻĽַݡ
	
	--11.clkʱƵʲܳSDRAMĲοƵʡûǸıclkƵʣһҪע⣬
	--ǿSDRAMÿˢµļʱ䡣clkƵ100MHzʱSDRAMÿ64msˢ8192ΡƵʱСˣ
	--ˢµʱȻ涨ˢƵʣôSDRAMڲݽ𻵡ʱû
	--޸Refresh_PeriodֵSDRAMˢʡ

architecture bhv of sdram_ctrl is

	constant Refresh_Period : integer := 10#781#;--when Sclk is 100MHz,8192 Refresh cycles happens very 64ms
	constant tRP_period     : integer := 10#2#  ; --3 clock period 
	constant tRC_period     : integer := 10#11# ;--12 clock period
	constant tMRD_period    : integer := 10#1#  ; --2 clock period 
	constant tRCD_period    : integer := 10#1#  ; --3 clock period 
	
	signal Refresh_cnt       : std_logic_vector(9 downto 0);
	signal Refresh_INT       : std_logic;
	signal Refresh_INT_clear : std_logic;
	signal user_INT          : std_logic;
	
	signal user_rd0_wr1_buf      : std_logic;
	signal user_burst_length_buf : std_logic_vector((COL_WIDTH-1) downto 0);
	signal user_start_addr_buf   : std_logic_vector((BANK_WIDTH+ROW_WIDTH+COL_WIDTH-1) downto 0);
	
	type SDRAM_states is(
		Initialize,
		Wait_Interrupt,
		Response_Refresh_INT,
		Response_user_INT
	);
	signal SDRAM_state : SDRAM_states;
	
	signal Initialize_done           : std_logic;
	signal Response_Refresh_INT_done : std_logic;
	signal Response_user_INT_done    : std_logic;
	
	signal Delay_cnt    : std_logic_vector((COL_WIDTH-1) downto 0);
	signal cnt_overflow : std_logic;
	
	type Initialize_states is(
		Initialize_NOP_Command,
		Initialize_Precharge_All_Bank,
		Initialize_Precharge_tRP,
		Initialize_Auto_Refresh_Command1,
		Initialize_Auto_Refresh_tRC1,
		Initialize_Auto_Refresh_Command2,
		Initialize_Auto_Refresh_tRC2,
		Load_Mode_Register_Command,
		Load_Mode_Register_tMRD,
		Initialize_Finish
	);
	signal Initialize_state : Initialize_states;
	
	type Refresh_Response_states is(
		Refresh_NOP_Command,
		Refresh_Auto_Refresh_Command,
		Refresh_Auto_Refresh_tRC,
		Refresh_Finish
	);
	signal Refresh_Response_state : Refresh_Response_states;
	
	type user_Response_states is(
		user_Response_start_NOP,
		user_Response_Bank_activation,
		user_Response_tRCD,
		user_Response_Column_Selection,
		user_Response_wait_Precharge1,
		user_Response_wait_Precharge2,
		user_Response_end_NOP
	);
	signal user_Response_state : user_Response_states;
	
	signal Delay_set             : std_logic_vector((COL_WIDTH-1) downto 0);
	signal sdram_data_valid_buff : std_logic_vector(3 downto 0)            ;
	signal dqm_for_read_mask     : std_logic;
	signal dqm_for_write_mask    : std_logic;
	
	type SDRAM_Operations is(
		Device_Deselect,
		No_Operation,
		Mode_Register_Set,
		Bank_Active,
		Precharge_selected_Bank,
		Precharge_All_Banks,
		Read_data,
		Read_with_Autoprecharge,
		Write_data,
		Write_with_Autoprecharge,
		Burst_stop,
		Auto_Refresh,
		Self_Refresh_Entry,
		Self_Refresh_Exit,
		Precharge_Power_Down_Entry,
		Precharge_Power_Down_Exit,
		Clock_Suspend_Entry,
		Clock_Suspend_Exit
	);
	signal SDRAM_Operation : SDRAM_Operations;
	
	signal CS_RAS_CAS_WE_n : std_logic_vector(3 downto 0);
	
begin
  
  process(clk)--account for Refresh_INT
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          Refresh_cnt <= (others => '0');
       elsif Refresh_cnt /= Refresh_Period then
          Refresh_cnt <= Refresh_cnt + 1;
       else
          Refresh_cnt <= (others => '0');
       end if;
    end if;
  end process;
  
  process(clk)
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          Refresh_INT <= '0';
       elsif Refresh_cnt = Refresh_Period then
          Refresh_INT <= '1';
       elsif Refresh_INT_clear = '1' then
          Refresh_INT <= '0';
       end if;
    end if;
  end process;
  
  user_INT <= user_command_req;--A read/write command should be excuted when user_INT is High
  
  process(clk)
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          user_rd0_wr1_buf      <= '0'            ;
          user_burst_length_buf <= (others => '0');
          user_start_addr_buf   <= (others => '0');
       elsif user_command_gnt = '1' then
          user_rd0_wr1_buf      <= user_rd0_wr1     ;
          user_burst_length_buf <= user_burst_length;
          user_start_addr_buf   <= user_start_addr  ;
       end if;
    end if;
  end process;
  
  process(clk)--SDRAM state machine
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          SDRAM_state <= Initialize;
       else
          case SDRAM_state is
            when Initialize =>
              if Initialize_done = '1' then
                 SDRAM_state <= Wait_Interrupt;
              end if;
            when Wait_Interrupt =>
              if Refresh_INT = '1' then
                 SDRAM_state <= Response_Refresh_INT;
              elsif user_INT = '1' then
                 SDRAM_state <= Response_user_INT;
              end if;
            when Response_Refresh_INT =>
              if Response_Refresh_INT_done = '1' then
                 SDRAM_state <= Wait_Interrupt;
              end if;
            when Response_user_INT =>
              if Response_user_INT_done = '1' then
                 SDRAM_state <= Wait_Interrupt;
              end if;
            when others =>
              SDRAM_state <= Initialize;
          end case;
       end if;
    end if;
  end process;
  
  process(clk)--Initialize state machine
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          Initialize_state <= Initialize_NOP_Command;
       elsif SDRAM_state = Initialize then
          case Initialize_state is
            when Initialize_NOP_Command =>--A NOP Command before Precharge
              Initialize_state <= Initialize_Precharge_All_Bank;
            when Initialize_Precharge_All_Bank =>--Precharge Command,all banks
              Initialize_state <= Initialize_Precharge_tRP;
            when Initialize_Precharge_tRP =>--tRP period,with NOP Command
              if cnt_overflow = '1' then
                 Initialize_state <= Initialize_Auto_Refresh_Command1;
              end if;
            when Initialize_Auto_Refresh_Command1 =>--The Frist Auto Refresh Command
              Initialize_state <= Initialize_Auto_Refresh_tRC1;
            when Initialize_Auto_Refresh_tRC1 =>--tRC period,with NOP Command
              if cnt_overflow = '1' then
                 Initialize_state <= Initialize_Auto_Refresh_Command2;
              end if;
            when Initialize_Auto_Refresh_Command2 =>--The Second Auto Refresh Command
              Initialize_state <= Initialize_Auto_Refresh_tRC2;
            when Initialize_Auto_Refresh_tRC2 =>--tRC period,with NOP Command
              if cnt_overflow = '1' then
                 Initialize_state <= Load_Mode_Register_Command;
              end if;
            when Load_Mode_Register_Command =>--Load Mode Register
              Initialize_state <= Load_Mode_Register_tMRD;
            when Load_Mode_Register_tMRD =>--tMRD period,with NOP Command
              if cnt_overflow = '1' then
                 Initialize_state <= Initialize_Finish;
              end if;
            when Initialize_Finish => 
              null;
            when others =>
              Initialize_state <= Initialize_NOP_Command;
          end case;
       else
          Initialize_state <= Initialize_NOP_Command;
       end if;
    end if;
  end process;
  
  process(clk)
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          Initialize_done <= '0';
       elsif Initialize_state = Initialize_Finish then
          Initialize_done <= '1';
       else
          Initialize_done <= '0';
       end if;
    end if;
  end process;
  
  process(clk)--Refresh response state machine
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          Refresh_Response_state <= Refresh_NOP_Command;
       elsif SDRAM_state = Response_Refresh_INT then
          case Refresh_Response_state is
            when Refresh_NOP_Command =>--Exit Power-Down state(if it's in) with NOP Command
              Refresh_Response_state <= Refresh_Auto_Refresh_Command;
            when Refresh_Auto_Refresh_Command =>--Auto Refresh Command
              Refresh_Response_state <= Refresh_Auto_Refresh_tRC;
            when Refresh_Auto_Refresh_tRC => --tRC period,with NOP Command
              if cnt_overflow = '1' then
                 Refresh_Response_state <= Refresh_Finish;
              end if;
            when Refresh_Finish =>
              null;
            when others =>
              Refresh_Response_state <= Refresh_NOP_Command;
          end case;
       else
          Refresh_Response_state <= Refresh_NOP_Command;
       end if;
    end if;
  end process;
  
  process(clk)
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          Response_Refresh_INT_done <= '0';
       elsif Refresh_Response_state = Refresh_Auto_Refresh_tRC and cnt_overflow = '1' then
          Response_Refresh_INT_done <= '1';
       else
          Response_Refresh_INT_done <= '0';
       end if;
    end if;
  end process;
  
  Refresh_INT_clear <= '1' when Refresh_Response_state = Refresh_Auto_Refresh_Command else '0';
  
  process(clk)--user response state machine
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          user_Response_state <= user_Response_start_NOP;
       elsif SDRAM_state = Response_user_INT then
          case user_Response_state is
            when user_Response_start_NOP =>--Exit Power-Down state with NOP Command
              user_Response_state <= user_Response_Bank_activation;
            when user_Response_Bank_activation =>--Bank/Row activation
              user_Response_state <= user_Response_tRCD;
            when user_Response_tRCD =>--tRCD period,with NOP Command
              if cnt_overflow = '1' then
                 user_Response_state <= user_Response_Column_Selection;
              end if;
            when user_Response_Column_Selection =>--Column Selection
              user_Response_state <= user_Response_wait_Precharge1;
            when user_Response_wait_Precharge1 =>--Wait to Precharge
              if cnt_overflow = '1' then
                 if user_rd0_wr1_buf = '0' then--if is read busrt
                    user_Response_state <= user_Response_end_NOP;
                 else--write busrt
                    user_Response_state <= user_Response_wait_Precharge2;
                 end if;
              end if;
            when user_Response_wait_Precharge2 =>--This state is special for the precharge of busrt write
              user_Response_state <= user_Response_end_NOP;
            when user_Response_end_NOP =>
              null;
            when others =>
              user_Response_state <= user_Response_start_NOP;
          end case;
       else
          user_Response_state <= user_Response_start_NOP;
       end if;
    end if;
  end process;
  
  process(clk)
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          Response_user_INT_done <= '0';
       elsif user_Response_state = user_Response_end_NOP and SDRAM_state = Response_user_INT then
          Response_user_INT_done <= '1';
       else
          Response_user_INT_done <= '0';
       end if;
    end if;
  end process;
  
  process(clk)
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          user_command_gnt <= '0';
       else
          case SDRAM_state is
            when Wait_Interrupt =>
              if Refresh_INT = '0' and user_INT = '1' then
                 user_command_gnt <= '1';
              else
                 user_command_gnt <= '0';
              end if;
            when others =>
              user_command_gnt <= '0';
          end case;
       end if;
    end if;
  end process;
  
  process(clk)
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          Delay_set <= (others => '0');
       else
         case SDRAM_state is
           when Initialize =>
             case Initialize_state is
               when Initialize_NOP_Command =>--load tRP period
                 Delay_set <=  conv_std_logic_vector(tRP_period,COL_WIDTH);
               when Initialize_Precharge_tRP =>--load tRC period
                 if cnt_overflow = '1' then
                    Delay_set <=  conv_std_logic_vector(tRC_period,COL_WIDTH);
                 end if;
               when Initialize_Auto_Refresh_tRC1 =>--load tRC period
                 if cnt_overflow = '1' then
                    Delay_set <=  conv_std_logic_vector(tRC_period,COL_WIDTH);
                 end if;
               when Initialize_Auto_Refresh_tRC2 =>--load tMRD period
                 if cnt_overflow = '1' then
                    Delay_set <=  conv_std_logic_vector(tMRD_period,COL_WIDTH);
                 end if;
               when others => null;
             end case;
           when Response_Refresh_INT =>
             case Refresh_Response_state is
               when Refresh_NOP_Command =>--load tRC period
                 Delay_set <=  conv_std_logic_vector(tRC_period,COL_WIDTH);
               when others => null;
             end case;
           when Response_user_INT =>
             case user_Response_state is
               when user_Response_start_NOP =>--load tRCD period
                 Delay_set <= conv_std_logic_vector(tRCD_period,COL_WIDTH);
               when user_Response_tRCD =>--load user burst length period
                 if cnt_overflow = '1' then
                    Delay_set <=  user_burst_length_buf;
                 end if;
               when others => null;
             end case;
           when others => null;
         end case;
       end if;
    end if;
  end process;
  
  process(clk)--Delay_cnt
  begin
  if clk'event and clk = '1' then
     if rst_n = '0' then
       Delay_cnt <= (others => '0');
     else
        case SDRAM_state is
          when Initialize =>
            case Initialize_state is
              when Initialize_Precharge_All_Bank    | --cnt for tRP period
                   Initialize_Auto_Refresh_Command1 | --cnt for tRC period
                   Initialize_Auto_Refresh_Command2 | --cnt for tRC period
                   Load_Mode_Register_Command       =>--cnt for tMRD period
                     Delay_cnt <= Delay_cnt + '1';
              when Initialize_Precharge_tRP         | --cnt for tRP period
                   Initialize_Auto_Refresh_tRC1     | --cnt for tRC period
                   Initialize_Auto_Refresh_tRC2     | --cnt for tRC period
                   Load_Mode_Register_tMRD          =>--cnt for tMRD period
                     if cnt_overflow = '1' then
                       Delay_cnt <= (others => '0');
                     else
                       Delay_cnt <= Delay_cnt + '1';
                     end if;
               when others => null; 
            end case;
          when Response_Refresh_INT =>
            case Refresh_Response_state is
              when Refresh_Auto_Refresh_Command =>--cnt for tRC period
                Delay_cnt <= Delay_cnt + '1';
              when Refresh_Auto_Refresh_tRC     =>--cnt for tRC period
                if cnt_overflow = '1' then                
                  Delay_cnt <= (others => '0');            
                else                            
                  Delay_cnt <= Delay_cnt + '1';
                end if;
              when others => null;
            end case;
          when Response_user_INT =>
            case user_Response_state is
              when user_Response_Bank_activation  | --cnt for tRCD period
                   user_Response_Column_Selection =>--cnt for user burst length
                     Delay_cnt <= Delay_cnt + '1';
              when user_Response_tRCD             | --cnt for tRCD period
                   user_Response_wait_Precharge1  =>--cnt for user burst length
                     if cnt_overflow = '1' then                
                       Delay_cnt <= (others => '0');           
                     else                            
                       Delay_cnt <= Delay_cnt + '1';
                     end if;
              when others => null;
            end case;
          when others => null;
        end case;
      end if;
  end if;
  end process;

  process(clk)--cnt_overflow
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          cnt_overflow <= '0';
       elsif Delay_cnt = Delay_set then
          cnt_overflow <= '1';
       else
          cnt_overflow <= '0';
       end if;
    end if;
  end process;
  
  process(clk)--sdram_data_valid_buff
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          sdram_data_valid_buff <= (others => '0');
       elsif user_Response_state = user_Response_wait_Precharge1 and user_rd0_wr1_buf ='0' then
          sdram_data_valid_buff <= sdram_data_valid_buff(2 downto 0) & '1';
       else
          sdram_data_valid_buff <= sdram_data_valid_buff(2 downto 0) & '0';
       end if;
    end if;
  end process;
  
  process(clk)--sdram_data_valid
  begin
    if clk'event and clk = '1' then
       if SDRAM_MR(5 downto 4) = "11" then--CAS Latency = 3
          sdram_data_valid <= sdram_data_valid_buff(3);
       else--other CAS Latency, treated it as 2
          sdram_data_valid <= sdram_data_valid_buff(2);
       end if;
    end if;
  end process;
  
  process(clk)--sdram_rd_data
  begin
    if clk'event and clk = '1' then
       sdram_rd_data <= sdram_dq_in;
    end if;
  end process;
  
  process(clk)--Tx_fifo_rd_en
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          Tx_fifo_rd_en <= '0';
       elsif user_rd0_wr1_buf = '1' then
          case user_Response_state is
            when user_Response_tRCD =>
              if cnt_overflow = '1' then
                 Tx_fifo_rd_en <= '1';
              end if;
            when user_Response_Column_Selection | 
                 user_Response_wait_Precharge1  =>
                   if Delay_cnt = Delay_set then
                      Tx_fifo_rd_en <= '0';
                   end if;
            when others => null;
          end case;
       else
          Tx_fifo_rd_en <= '0';
       end if;
    end if;
  end process;
  
  process(clk)--sdram_dq_out_en
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          sdram_dq_out_en <= '0';
       else
          sdram_dq_out_en <= Tx_fifo_rd_en;
       end if;
    end if;
  end process;
  
  sdram_dq_out <= Tx_fifo_rd_data((DQ_WIDTH-1) downto 0);
  
  process(clk)--dqm_for_read_mask
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          dqm_for_read_mask <= '0';
       else
	      case user_Response_state is
	        when user_Response_Column_Selection =>
	          if user_rd0_wr1_buf = '0' then
	             dqm_for_read_mask <= '1';
	          end if;
	        when user_Response_end_NOP =>
	        	dqm_for_read_mask <= '0';
	        when others => null;
	      end case;
       end if;
    end if;
  end process;
  
  process(clk)--dqm_for_write_mask
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          dqm_for_write_mask <= '0';
       else
	      dqm_for_write_mask <= Tx_fifo_rd_en;
       end if;
    end if;
  end process;
  
  process(dqm_for_read_mask,dqm_for_write_mask,Tx_fifo_rd_data)
  begin
    if dqm_for_read_mask = '1' then
       sdram_dqm <= (others => '0');
    elsif dqm_for_write_mask = '1' then
       sdram_dqm <= Tx_fifo_rd_data((DQ_WIDTH/8+DQ_WIDTH-1) downto DQ_WIDTH);
    else
       sdram_dqm <= (others => '1');
    end if;
  end process;
  
  process(clk)
  begin
    if clk'event and clk = '1' then
       if rst_n = '0' then
          sdram_ba <= (others => '0');
          sdram_addr <= (others => '0');
          sdram_cke <= '0';
          SDRAM_Operation <= Device_Deselect;
       else
          case SDRAM_state is
            when Initialize =>
              case Initialize_state is
                when Initialize_NOP_Command =>--this operation can be replaced by Precharge_Power_Down_Exit
                  sdram_cke <= '1';
                  SDRAM_Operation <= No_Operation;
                when Initialize_Precharge_All_Bank =>
                  sdram_addr(10) <= '1';
                  sdram_cke <= '1';
                  SDRAM_Operation <= Precharge_All_Banks;
                when Initialize_Precharge_tRP =>
                  sdram_cke <= '1';
                  SDRAM_Operation <= No_Operation;
                when Initialize_Auto_Refresh_Command1 =>
                  sdram_cke <= '1';
                  SDRAM_Operation <= Auto_Refresh;
                when Initialize_Auto_Refresh_tRC1 =>
                  sdram_cke <= '1';
                  SDRAM_Operation <= No_Operation;
                when Initialize_Auto_Refresh_Command2 =>
                  sdram_cke <= '1';
                  SDRAM_Operation <= Auto_Refresh;
                when Initialize_Auto_Refresh_tRC2 =>
                  sdram_cke <= '1';
                  SDRAM_Operation <= No_Operation;
                when Load_Mode_Register_Command =>
                  sdram_ba <= "00";
                  sdram_addr <= SDRAM_MR((ROW_WIDTH-1) downto 0);
                  sdram_cke <= '1';
                  SDRAM_Operation <= Mode_Register_Set;
                when Load_Mode_Register_tMRD =>
                  sdram_cke <= '1';
                  SDRAM_Operation <= No_Operation;
                when others => null;
              end case;
            when Wait_Interrupt =>--this operation can be replaced by Precharge_Power_Down_Entry
              sdram_cke <= '0';
              SDRAM_Operation <= No_Operation;
            when Response_Refresh_INT =>
              case Refresh_Response_state is
                when Refresh_NOP_Command =>--this operation can be replaced by Precharge_Power_Down_Exit
                  sdram_cke <= '1';
                  SDRAM_Operation <= No_Operation;
                when Refresh_Auto_Refresh_Command =>
                  sdram_cke <= '1';
                  SDRAM_Operation <= Auto_Refresh;
                when Refresh_Auto_Refresh_tRC =>
                  sdram_cke <= '1';
                  SDRAM_Operation <= No_Operation;
                when others => null;
              end case;
            when Response_user_INT =>
              case user_Response_state is
                when user_Response_start_NOP =>--this operation can be replaced by Precharge_Power_Down_Exit
                  sdram_cke <= '1';
                  SDRAM_Operation <= No_Operation;
                when user_Response_Bank_activation =>
                  sdram_ba <= user_start_addr_buf((BANK_WIDTH+ROW_WIDTH+COL_WIDTH-1) downto (ROW_WIDTH+COL_WIDTH));
                  sdram_addr <= user_start_addr_buf((ROW_WIDTH+COL_WIDTH-1) downto COL_WIDTH);
                  sdram_cke <= '1';
                  SDRAM_Operation <= Bank_Active;
                when user_Response_tRCD =>
                  sdram_cke <= '1';
                  SDRAM_Operation <= No_Operation;
                when user_Response_Column_Selection =>
                  sdram_addr((COL_WIDTH-1) downto 0) <= user_start_addr_buf((COL_WIDTH-1) downto 0);
                  sdram_addr(10) <= '0';
                  sdram_cke <= '1';
                  if user_rd0_wr1_buf = '1' then
                     SDRAM_Operation <= Write_data;
                  else
                     SDRAM_Operation <= Read_data;
                  end if;
                when user_Response_wait_Precharge1 =>
                  sdram_cke <= '1';
                  if cnt_overflow = '1' and user_rd0_wr1_buf = '0' then--only generate precharge cmd when busrt read
                     sdram_addr(10) <= '1';
                     SDRAM_Operation <= Precharge_All_Banks;--this behavor can be replaced by Precharge_selected_Bank
                  else
                     SDRAM_Operation <= No_Operation;
                  end if;
                when user_Response_wait_Precharge2 =>
                  sdram_addr(10) <= '1';
                  SDRAM_Operation <= Precharge_All_Banks;--this behavor can be replaced by Precharge_selected_Bank
                when user_Response_end_NOP =>
                  sdram_cke <= '1';
                  SDRAM_Operation <= No_Operation;
                when others => null;
              end case;
            when others =>
              sdram_cke <= '1';
              SDRAM_Operation <= No_Operation;
          end case;
       end if;
    end if;
  end process;
  
  process(SDRAM_Operation)--sdram_cs_n,sdram_ras_n,sdram_cas_n,sdram_we_n
  begin
    case SDRAM_Operation is
      when Device_Deselect =>--CKE should be HIGH at the last clock period
        CS_RAS_CAS_WE_n <= "1111";
      when No_Operation =>--CKE should be HIGH at the last clock period
        CS_RAS_CAS_WE_n <= "0111";
      when Mode_Register_Set =>--CKE should be HIGH at the last clock period
        CS_RAS_CAS_WE_n <= "0000";
      when Bank_Active =>--CKE should be HIGH at the last clock period
        CS_RAS_CAS_WE_n <= "0011";
      when Precharge_selected_Bank =>--CKE should be HIGH at the last clock period
        CS_RAS_CAS_WE_n <= "0010";
      when Precharge_All_Banks =>--CKE should be HIGH at the last clock period
        CS_RAS_CAS_WE_n <= "0010";
      when Read_data =>--CKE should be HIGH at the last clock period
        CS_RAS_CAS_WE_n <= "0101";
      when Read_with_Autoprecharge =>--CKE should be HIGH at the last clock period
        CS_RAS_CAS_WE_n <= "0101";
      when Write_data =>--CKE should be HIGH at the last clock period
        CS_RAS_CAS_WE_n <= "0100";
      when Write_with_Autoprecharge =>--CKE should be HIGH at the last clock period
        CS_RAS_CAS_WE_n <= "0100";
      when Burst_stop =>--CKE should be HIGH at the last clock period
        CS_RAS_CAS_WE_n <= "0110";
      when Auto_Refresh =>--CKE should be HIGH at the last clock period
        CS_RAS_CAS_WE_n <= "0001";
      when Self_Refresh_Entry =>--CKE should be HIGH at the last clock period
        CS_RAS_CAS_WE_n <= "0001";
      when Self_Refresh_Exit =>--CKE should be LOW at the last clock period
        CS_RAS_CAS_WE_n <= "0111";
      when Precharge_Power_Down_Entry =>--CKE should be HIGH at the last clock period
        CS_RAS_CAS_WE_n <= "0111";
      when Precharge_Power_Down_Exit =>--CKE should be LOW at the last clock period
        CS_RAS_CAS_WE_n <= "0111";
      when Clock_Suspend_Entry =>--CKE should be HIGH at the last clock period
        CS_RAS_CAS_WE_n <= "1111";
      when Clock_Suspend_Exit =>--CKE should be LOW at the last clock period
        CS_RAS_CAS_WE_n <= "1111";
      when others =>--treat it as No_Operation
        CS_RAS_CAS_WE_n <= "0111";
    end case;
  end process;
  
  sdram_cs_n  <= CS_RAS_CAS_WE_n(3);
  sdram_ras_n <= CS_RAS_CAS_WE_n(2);
  sdram_cas_n <= CS_RAS_CAS_WE_n(1);
  sdram_we_n  <= CS_RAS_CAS_WE_n(0);
  
end bhv;