library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity sdram_io is
generic(
	ROW_WIDTH  : integer := 12;--2^12 = 4096 addresses in each bank
	BANK_WIDTH : integer := 2 ;--2^2 = 4 banks in a single chip
	DQ_WIDTH   : integer := 8  --8 bit Data Bus for one chip
);
port(
	clk0 : in std_logic;
	clk1 : in std_logic;
	
	c0_sdram_cke       : in  std_logic                                ;
	c0_sdram_cs_n      : in  std_logic                                ;
	c0_sdram_ras_n     : in  std_logic                                ;
	c0_sdram_cas_n     : in  std_logic                                ;
	c0_sdram_we_n      : in  std_logic                                ;
	c0_sdram_ba        : in  std_logic_vector((BANK_WIDTH-1) downto 0);
	c0_sdram_addr      : in  std_logic_vector((ROW_WIDTH-1) downto 0) ;
	c0_sdram_dq_out_en : in  std_logic                                ;
	c0_sdram_dqm       : in  std_logic_vector((DQ_WIDTH/8-1) downto 0);
	c0_sdram_dq_out    : in  std_logic_vector((DQ_WIDTH-1)   downto 0);
	c0_sdram_dq_in     : out std_logic_vector((DQ_WIDTH-1)   downto 0);
	
	c1_sdram_clk     : out   std_logic                                ;
	c1_sdram_cke     : out   std_logic                                ;
	c1_sdram_cs_n    : out   std_logic                                ;
	c1_sdram_ras_n   : out   std_logic                                ;
	c1_sdram_cas_n   : out   std_logic                                ;
	c1_sdram_we_n    : out   std_logic                                ;
	c1_sdram_ba      : out   std_logic_vector((BANK_WIDTH-1) downto 0);
	c1_sdram_addr    : out   std_logic_vector((ROW_WIDTH-1)  downto 0);
	c1_sdram_dqm     : out   std_logic_vector((DQ_WIDTH/8-1) downto 0);
	c1_sdram_dq      : inout std_logic_vector((DQ_WIDTH-1)   downto 0) 
);
end sdram_io;

architecture bhv of sdram_io is
	
	signal c1_sdram_dq_out    : std_logic_vector((DQ_WIDTH-1)   downto 0);
	signal c1_sdram_dq_out_en : std_logic                                ;
	
begin
	
	process(clk0)
	begin
		if clk0'event and clk0 = '1' then
			c1_sdram_cke       <= c0_sdram_cke      ;
			c1_sdram_cs_n      <= c0_sdram_cs_n     ;
			c1_sdram_ras_n     <= c0_sdram_ras_n    ;
			c1_sdram_cas_n     <= c0_sdram_cas_n    ;
			c1_sdram_we_n      <= c0_sdram_we_n     ;
			c1_sdram_ba        <= c0_sdram_ba       ;
			c1_sdram_addr      <= c0_sdram_addr     ;
			c1_sdram_dqm       <= c0_sdram_dqm      ;
			c1_sdram_dq_out    <= c0_sdram_dq_out   ;
			c1_sdram_dq_out_en <= c0_sdram_dq_out_en;
		end if; 
	end process;
	
	c1_sdram_dq <= c1_sdram_dq_out when c1_sdram_dq_out_en = '1' else (others => 'Z');
	c0_sdram_dq_in <= c1_sdram_dq;
	c1_sdram_clk <= clk1;

end bhv; 